import dask.array as da
import numpy as np
from typing import Union, Tuple, List, Sequence
from mrcfile.mrcmemmap import MrcMemmap
from pathlib import Path
from dask.array.core import normalize_chunks
from numpy.typing import ArrayLike
from xarray import DataArray

Pathlike = Union[str, Path]


def access_mrc(path: Pathlike, mode: str, **kwargs):
    return MrcMemmap(path, mode=mode, **kwargs)


def mrc_shape_dtype_inference(mem: MrcMemmap) -> Tuple[Tuple[int], str]:
    """
    Infer the shape and datatype of an MrcMemmap array. We cannot use the dtype attribute
    because the MRC2014 specification does not officially support the uint8 datatype,
    but that doesn't stop people from storing uint8 data as int8. This can only be inferred
    by checking if the header.dmax propert exceeds the upper limit of int8 (127).
    """
    shape = mem.data.shape
    dtype = mem.data.dtype
    if dtype == "int8":
        if mem.header.dmax > 127:
            dtype = "uint8"
    return shape, dtype


def mrc_coordinate_inference(mem: MrcMemmap) -> List[DataArray]:
    header = mem.header
    grid_size_angstroms = header.cella
    coords = []
     
    if mem.data.flags['C_CONTIGUOUS']:
        # we reverse the keys from (x,y,z) to (z,y,x) so the order matches
        # numpy indexing order
        keys = reversed(header.cella.dtype.fields.keys())
    else: 
        keys = header.cella.dtype.fields.keys()
    for key in keys:
        axis = np.linspace(header[f'n{key}start'], (grid_size_angstroms[key] / 10), header[f'n{key}'])
        coords.append(DataArray(data=axis, dims=(key,), attrs={'units': 'nm'}))
    
    return coords 


def mrc_chunk_loader(fname, block_info=None):
    idx = tuple(slice(*idcs) for idcs in block_info[None]["array-location"])
    # block_info[None] contains the output specification 
    dtype = block_info[None]['dtype']
    result = np.array(access_mrc(fname, mode="r").data[idx]).astype(dtype)
    return result


def mrc_to_dask(urlpath: Pathlike, chunks: Union[str, Sequence[int]]):
    """
    Generate a dask array backed by a memory-mapped .mrc file
    """
    with access_mrc(urlpath, mode="r") as mem:
        shape, dtype = mrc_shape_dtype_inference(mem)

    if chunks=='auto':
        _chunks = normalize_chunks((1, *(-1,) * (len(shape) -1)), shape, dtype=dtype)
    else:
        _chunks = normalize_chunks(chunks, shape, dtype=dtype)

    arr = da.map_blocks(mrc_chunk_loader, urlpath, chunks=_chunks, dtype=dtype)
    return arr



from .fibsem import read_fibsem
from pathlib import Path
from typing import (
    Union,
    Iterable,
    List,
    Optional,
    Callable,
    Dict,
    Tuple,
    Sequence,
    Any,
)
from dask import delayed
import dask.array as da
import os
from itertools import groupby
from collections import defaultdict
import zarr
import h5py
import mrcfile
from xarray import DataArray
import numpy as np
from dask import bag
from .mrc import mrc_coordinate_inference, mrc_shape_dtype_inference, access_mrc, mrc_to_dask
from .util import split_path_at_suffix
from .zarr import zarr_array_from_dask, access_n5, access_zarr, n5_to_dask, zarr_n5_coordinate_inference, zarr_to_dask
from .tensorstore import access_precomputed, precomputed_to_dask
from numcodecs import GZip
import fsspec
import toolz as tz
from glob import glob
import distributed

# encode the fact that the first axis in zarr is the z axis
_zarr_axes = {"z": 0, "y": 1, "x": 2}
# encode the fact that the first axis in n5 is the x axis
_n5_axes = {"z": 2, "y": 1, "x": 0}
_formats = (".dat", ".mrc")
_container_extensions = (".zarr", ".n5", ".h5", ".precomputed")
_suffixes = (*_formats, *_container_extensions)

Pathlike = Union[str, Path]
defaultUnit = "nm"


def broadcast_kwargs(**kwargs) -> Dict:
    """
    For each keyword: arg in kwargs, assert that there are only 2 types of args: sequences with length = 1
    or sequences with some length = k. Every arg with length 1 will be repeated k times, such that the return value
    is a dict of kwargs with minimum length = k.
    """
    grouped: Dict[str, List] = defaultdict(list)
    sorter = lambda v: len(v[1])
    s = sorted(kwargs.items(), key=sorter)
    for l, v in groupby(s, key=sorter):
        grouped[l].extend(v)

    assert len(grouped.keys()) <= 2
    if len(grouped.keys()) == 2:
        assert min(grouped.keys()) == 1
        output_length = max(grouped.keys())
        singletons, nonsingletons = tuple(grouped.values())
        singletons = ((k, v * output_length) for k, v in singletons)
        result = {**dict(singletons), **dict(nonsingletons)}
    else:
        result = kwargs

    return result


def access_fibsem(path: Union[Pathlike, Iterable[str], Iterable[Path]], mode: str):
    if mode != "r":
        raise ValueError(
            f".dat files can only be accessed in read-only mode, not {mode}."
        )
    return read_fibsem(path)


def access_h5(
    dir_path: Pathlike, container_path: Pathlike, mode: str, **kwargs
) -> Union[h5py.Dataset, h5py.Group]:
    result = h5py.File(dir_path, mode=mode, **kwargs)
    if container_path != "":
        result = result[str(container_path)]
    return result


accessors: Dict[str, Callable[..., Any]] = {}
accessors[".dat"] = access_fibsem
accessors[".n5"] = access_n5
accessors[".zarr"] = access_zarr
accessors[".h5"] = access_h5
accessors[".mrc"] = access_mrc
accessors[".precomputed"] = access_precomputed

daskifiers: Dict[str, Callable[..., da.core.Array]] = {}
daskifiers[".mrc"] = mrc_to_dask
daskifiers[".n5"] = n5_to_dask
daskifiers[".zarr"] = zarr_to_dask
daskifiers[".precomputed"] = precomputed_to_dask


def access(
    path: Union[Pathlike, Iterable[str], Iterable[Path]],
    mode: str,
    **kwargs,
) -> Any:
    """

    Access data from a variety of array storage formats.

    Parameters
    ----------
    path: A path or collection of paths to image files. If `path` is a string, then the appropriate reader will be
          selected based on the extension of the path, and the file will be read. If `path` is a collection of strings,
          it is assumed that each string is a path to an image and each will be read sequentially.

    lazy: A boolean, defaults to False. If True, this function returns the native file reader wrapped by
    dask.delayed. This is advantageous for distributed computing.

    mode: The access mode for the file. e.g. 'r' for read-only access.

    Returns an array-like object, a collection of array-like objects, a chunked store, or
    a dask.delayed object.
    -------

    """
    if isinstance(path, (str, Path)):
        path_outer, path_inner, suffix = split_path_at_suffix(path, _suffixes)
        is_container = suffix in _container_extensions

        try:
            accessor = accessors[suffix]
        except KeyError:
            raise ValueError(
                f"Cannot access images with extension {suffix}. Try one of {list(accessors.keys())}"
            )

        if is_container:
            return accessor(path_outer, path_inner, mode=mode, **kwargs)
        else:
            return accessor(path_outer, mode=mode, **kwargs)

    elif isinstance(path, Iterable):
        return [access(p, mode, **kwargs) for p in path]
    else:
        raise ValueError("`path` must be a string or iterable of strings")


def read(path: Union[Pathlike, Iterable[str], Iterable[Path]], **kwargs):
    """

    Access data on disk with read-only permissions

    Parameters
    ----------
    path: A path or collection of paths to image files. If `path` is a string, then the appropriate image reader will be
          selected based on the extension of the path, and the file will be read. If `path` is a collection of strings,
          it is assumed that each string is a path to an image and each will be read sequentially.

    lazy: A boolean, defaults to False. If True, this function returns the native file reader wrapped by
    dask.delayed. This is advantageous for distributed computing.

    Returns an array-like object, a collection of array-like objects, a chunked store, or
    a dask.delayed object.
    -------

    """
    return access(path, mode="r", **kwargs)


def read_dask(urlpath: str, chunks='auto', **kwargs) -> da.core.Array:
    """
    Create a dask array from a path
    """
    path_outer, path_inner, suffix = split_path_at_suffix(urlpath, _suffixes)
    return daskifiers[suffix](urlpath, chunks, **kwargs)


def read_xarray(urlpath: str, chunks: Union[str, Tuple[int,...]]='auto', coords: Any='auto', **kwargs) -> DataArray:
    """
    Create an xarray.DataArray from data found at a path. 
    """
    raw_array = read(urlpath)
    if hasattr(raw_array, 'attrs'):
        if not kwargs.get('attrs'):
            kwargs.update({'attrs': dict(raw_array.attrs)}) 
    if kwargs.get('attrs'):
        kwargs['attrs'].update({'urlpath': urlpath})
    dask_array = read_dask(urlpath, chunks=chunks)
    if coords == 'auto':
        coords = infer_coordinates(raw_array)    
    result = DataArray(dask_array, coords=coords, **kwargs)
    return result


def infer_coordinates(arr: Any, default_unit: str="nm") -> List[DataArray]:
    if isinstance(arr, zarr.core.Array):
        coords = zarr_n5_coordinate_inference(arr)
    elif isinstance(arr, mrcfile.mrcmemmap.MrcMemmap):
        coords = mrc_coordinate_inference(arr)
    else:
        raise ValueError(f'No coordinate inference possible for array of type {type(arr)}')
    return coords


def DataArrayFactory(arr: Any, **kwargs) -> DataArray:
    """
    Create an xarray.DataArray from an array-like input (e.g., zarr array, dask array). This is a very light
    wrapper around the xarray.DataArray constructor that checks for cosem/n5 metadata attributes and uses those to
    generate DataArray.coords and DataArray.dims properties; additionally, metadata about units will be inferred and
    inserted into the `attrs` kwarg if it is supplied.

    Parameters
    ----------

    arr: Array-like object (dask array or zarr array)

    """
    attrs: Optional[Dict]
    extra_attrs = {}

    # if we pass in a zarr array, daskify it first
    # maybe later add hdf5 support here
    if isinstance(arr, zarr.core.Array):
        source = str(Path(arr.store.path) / arr.path)
        # save the full path to the array as an attribute
        extra_attrs["source"] = source
        arr = da.from_array(arr, chunks=arr.chunks)

    if attrs := kwargs.get("attrs"):
        out_attrs = attrs.copy()
        out_attrs.update(extra_attrs)
        kwargs["attrs"] = out_attrs
    else:
        kwargs["attrs"] = extra_attrs

    data = DataArray(arr, **kwargs)
    return data


def populate_group(
    container_path: Pathlike,
    group_path: Pathlike,
    arrays: Sequence[DataArray],
    array_paths: Sequence[str],
    chunks: Sequence[int],
    group_attrs: Dict[str, Any] = {},
    compressor: Any = GZip(-1),
    array_attrs: Optional[Sequence[Dict[str, Any]]] = None,
) -> Tuple[zarr.hierarchy.group, Tuple[zarr.core.Array]]:

    zgroup = access(
        os.path.join(container_path, group_path), mode="w", attrs=group_attrs
    )
    zarrays = []
    if array_attrs == None:
        _array_attrs = ({},) * len(arrays)
    else:
        _array_attrs = array_attrs

    for idx, arr in enumerate(arrays):
        path = os.path.join(container_path, group_path, array_paths[idx])
        chunking = chunks[idx]
        compressor = compressor
        attrs = _array_attrs[idx]
        zarrays.append(
            access(
                path,
                shape=arr.shape,
                dtype=arr.dtype,
                chunks=chunking,
                compressor=compressor,
                attrs=attrs,
                mode="w",
            )
        )
    return zgroup, zarrays


def infer_dtype(path: str) -> str:
    fd = read(path)
    if hasattr(fd, "dtype"):
        dtype = str(fd.dtype)
    elif hasattr(fd, "data"):
        _, dtype = mrc_shape_dtype_inference(fd)
        dtype = str(dtype)
    else:
        raise ValueError(f"Cannot infer dtype of data located at {path}")
    return dtype


def sequential_rechunk(source: Any, target: Any, slab_size: Tuple[int], intermediate_chunks: Tuple[int], client: distributed.Client, num_workers: int) -> List[None]:
    """
    Load slabs of an array into local memory, then create a dask array and rechunk that dask array, then store into 
    chunked array storage.
    """
    results = []
    slices = da.core.slices_from_chunks(source.rechunk(slab_size).chunks)
    
    for sl in slices:
        arr_in = source[sl].compute(scheduler='threads')
        darr_in = da.from_array(arr_in, chunks=intermediate_chunks)
        store_op = da.store(darr_in,target, regions=sl, compute=False, lock=None)
        client.cluster.scale(num_workers)
        results.extend(client.compute(store_op).result())
        client.cluster.scale(0)
    return results


from dask import delayed, bag
from shutil import rmtree
from glob import glob
import os
from typing import List
import toolz as tz
from typing import Sequence, Union, Any, Tuple
from pathlib import Path
import fsspec


@delayed
def _rmtree_after_delete_files(path: str, dependency: Any):
    rmtree(path)


def rmtree_parallel(
    path: Union[str, Path], branch_depth: int = 1, compute: bool = True
):
    branches = glob(os.path.join(path, *("*",) * branch_depth))
    deleter = os.remove
    files = list_files_parallel(branches)
    deleted_files = bag.from_sequence(files).map(deleter)
    result = _rmtree_after_delete_files(path, dependency=deleted_files)

    if compute:
        return result.compute(scheduler="threads")
    else:
        return result


def list_files(paths: Union[Sequence[Union[str, Path]], str, Path]):
    if isinstance(paths, str) or isinstance(paths, Path):
        if os.path.isdir(paths):
            return list(
                tz.concat(
                    (os.path.join(dp, f) for f in fn) for dp, dn, fn in os.walk(paths)
                )
            )
        elif os.path.isfile(paths):
            return [paths]
        else:
            raise ValueError(f"Input argument {paths} is not a path or a directory")

    elif isinstance(paths, Sequence):
        sortd = sorted(paths, key=os.path.isdir)
        files, dirs = tuple(tz.partitionby(os.path.isdir, sortd))
        return list(tz.concatv(files, *tz.map(list_files, dirs)))


def list_files_parallel(
    paths: Union[Sequence[Union[str, Path]], str, Path], compute: bool = True
):
    result = []
    delf = delayed(list_files)

    if isinstance(paths, str) or isinstance(paths, Path):
        result = bag.from_delayed([delf(paths)])
    elif isinstance(paths, Sequence):
        result = bag.from_delayed([delf(p) for p in paths])
    else:
        raise TypeError(f"Input must be a string or a sequence, not {type(paths)}")

    if compute:
        return result.compute(scheduler="threads")
    else:
        return result


def split_path_at_suffix(urlpath: str, suffixes: Sequence[str]) -> Tuple[str, str, str]:
    """
    Given a string representing a path on a filesystem and a collection of suffixes, return
    the path split at the last instance of any element of the path containing one of the
    suffixes, as well as the suffix. If the last element of the path bears a suffix, return the path,
    the empty string, and the suffix.
    """
    protocol: str 
    subpath: str
    protocol, subpath = fsspec.core.split_protocol(urlpath)
    parts = Path(subpath).parts
    suffixed = [Path(part).suffix in suffixes for part in parts]

    if not any(suffixed):
        raise ValueError("No suffixed path elements found")

    index = [idx for idx, val in enumerate(suffixed) if val][-1]
    if index == (len(parts) - 1):
        pre, post = subpath, ""
    else:
        pre, post = os.path.join(*parts[: index + 1]), os.path.join(*parts[index + 1 :])

    suffix = Path(pre).suffix
    if protocol:
        pre = f"{protocol}://{pre}"
    return pre, post, suffix



from dataclasses import dataclass
from typing import Any, Dict, Optional, Sequence, Union, List
import tensorstore as ts
import json
from pathlib import Path
from xarray.core.dataarray import DataArray
from dacite import from_dict
import fsspec
import os
from dask.array.core import normalize_chunks
from dask.array import map_blocks

from fibsem_tools.io.util import split_path_at_suffix

DRIVERS = {"n5", "neuroglancer_precomputed"}
KVSTORE_DRIVERS = {"file", "gcs"}


@dataclass
class StripNullFields:
    def asdict(self):
        result = {}
        for k, v in self.__dict__.items():
            if v is not None:
                if hasattr(v, "asdict"):
                    result[k] = v.asdict()
                elif isinstance(v, list):
                    result[k] = []
                    for element in v:
                        if hasattr(element, "asdict"):
                            result[k].append(element.asdict())
                        else:
                            result[k].append(element)
                else:
                    result[k] = v
        return result


@dataclass
class KVStore(StripNullFields):
    driver: str
    path: str


@dataclass
class Sharding(StripNullFields):
    preshift_bits: int
    hash: str
    minishard_bits: int
    shard_bits: int
    minishard_index_encoding: str = "raw"
    data_index_encoding: str = "raw"
    at_type: str = "neuroglancer_uint64_sharded_v1"


@dataclass
class ScaleMetadata(StripNullFields):
    size: Optional[Sequence[int]]
    encoding: Optional[str]
    chunk_size: Optional[List[int]]
    resolution: Optional[Sequence[float]]
    key: Optional[str] = None
    voxel_offset: Optional[Sequence[int]] = None
    jpeg_quality: Optional[int] = None
    sharding: Optional[Sharding] = None

    def __post_init__(self):
        if self.resolution:
            self.resolution = [float(r) for r in self.resolution]
            if self.key == None:
                self.key = "_".join(map(str, self.resolution))


@dataclass
class MultiscaleMetadata(StripNullFields):
    type: str
    data_type: str
    num_channels: int


@dataclass
class TensorStoreSpec(StripNullFields):
    driver: str
    kvstore: KVStore
    path: str
    scale_index: Optional[int] = 0
    scale_metadata: Optional[ScaleMetadata] = None
    multiscale_metadata: Optional[MultiscaleMetadata] = None


@dataclass
class PrecomputedMetadata(StripNullFields):
    type: str
    data_type: str
    num_channels: int
    scales: Sequence[ScaleMetadata]
    at_type: Optional[str] = None

    @staticmethod
    def from_json(json) -> "PrecomputedMetadata":
        _json = json.copy()
        if "@type" in _json:
            at_type = _json.pop("@type")
            _json["at_type"] = at_type
        return from_dict(PrecomputedMetadata, _json)

    def __post_init__(self):
        _scales = []
        for scale in self.scales:
            if isinstance(scale, dict):
                _scales.append(ScaleMetadata(**scale))
            else:
                _scales.append(scale)
        self.scales = _scales


def parse_info(json_data: Dict[str, Any]) -> PrecomputedMetadata:
    at_type = json_data["@type"]
    volume_type = json_data["type"]
    data_type = json_data["data_type"]
    num_channels = json_data["num_channels"]
    scales = []
    for scale in json_data["scales"]:
        scale_copy = dict(**scale)
        chunk_sizes = scale_copy.pop("chunk_sizes")
        # take the first element from chunk_sizes
        scale_copy["chunk_size"] = chunk_sizes[0]
        scales.append(ScaleMetadata(**scale_copy))

    return PrecomputedMetadata(
        at_type=at_type,
        type=volume_type,
        data_type=data_type,
        num_channels=num_channels,
        scales=scales,
    )


def PrecomputedFromDataArray(
    dataarray,
    path,
    encoding,
    volume_type,
    num_channels=1,
    voxel_offset=None,
    key=None,
    chunk_size=None,
    **kwargs,
):
    assert len(dataarray.coords) == dataarray.ndim
    if voxel_offset is None:
        voxel_offset = (0,) * dataarray.ndim

    resolution = [
        abs(float(a[1].values - a[0].values)) for a in dataarray.coords.values()
    ]
    tsa = TensorStoreArray(
        driver="neuroglancer_precomputed",
        path=Path(path).parts[-1],
        kvstore_driver="file",
        kvstore_path=str(Path(path).parent),
        encoding=encoding,
        volume_type=volume_type,
        num_channels=num_channels,
        voxel_offset=voxel_offset,
        resolution=resolution,
        key=key,
        template=dataarray,
        chunk_size=chunk_size,
    )
    return tsa.open(**kwargs).result()


# todo: formally distinguish arrays from groups/containers
class TensorStoreArray:
    def __init__(
        self,
        driver,
        path,
        kvstore_driver,
        kvstore_path,
        key,
        encoding=None,
        num_channels=None,
        volume_type=None,
        resolution=None,
        jpeg_quality=None,
        voxel_offset=None,
        dtype=None,
        size=None,
        template=None,
        chunk_size=None,
        sharding=None,
        scale_index=None,
    ):
        """
        A (relatively) friendly interface to tensorstore arrays.
        """

        if template is not None:
            if size is not None:
                raise ValueError("Must supply either size or template but not both.")
            if dtype is not None:
                raise ValueError("Must supply either dtype or template but not both.")

            size = template.shape
            dtype = template.dtype.name
            if chunk_size is None:
                if hasattr(template, "chunks"):
                    chunk_size = [c[0] for c in template.chunks]
                else:
                    raise ValueError("Required argument `chunk_size` not supplied.")

            if dtype != "uint8" and encoding == "jpeg":
                raise ValueError(
                    f"JPEG encoding only works for uint8 arrays. You specified a dtype of {dtype}"
                )

        scale_metadata = ScaleMetadata(
            chunk_size=chunk_size,
            size=size,
            voxel_offset=voxel_offset,
            resolution=resolution,
            key=key,
            jpeg_quality=jpeg_quality,
            encoding=encoding,
            sharding=sharding,
        )

        kvstore = KVStore(driver=kvstore_driver, path=kvstore_path)

        multiscale_metadata = MultiscaleMetadata(
            data_type=dtype, num_channels=num_channels, type=volume_type
        )

        self.spec = TensorStoreSpec(
            driver,
            kvstore,
            path,
            scale_index=scale_index,
            scale_metadata=scale_metadata,
            multiscale_metadata=multiscale_metadata,
        ).asdict()

    def __repr__(self) -> str:
        return str(self.spec)

    def open(self, **kwargs):
        return ts.open(self.spec, **kwargs)


def access_precomputed(
    store_path: str,
    key: str,
    mode: str,
    array_type=None,
    dtype=None,
    num_channels=None,
    shape=None,
    resolution=None,
    encoding=None,
    chunks=None,
    jpeg_quality=None,
    voxel_offset=None,
    scale_index=None,
) -> TensorStoreArray:
    driver = "neuroglancer_precomputed"
    
    kvstore_driver, _store_path = fsspec.core.split_protocol(store_path)
    if kvstore_driver == None:
        kvstore_driver = "file"
        kvstore_path = "/"
        # remove the leading slash after making the absolute path
        _store_path = os.path.abspath(_store_path)[1:]
    else:
        kvstore_path = _store_path.split(os.path.sep)[0]

    if kvstore_driver not in KVSTORE_DRIVERS:
        raise ValueError(
            f"File system protocol {kvstore_driver} is not supported by tensorstore."
        )

    info_path = os.path.join(store_path, "info")

    if mode == 'r':
        with fsspec.open(info_path) as fh:
            json_data = json.loads(fh.read())
            precomputed_metadata = parse_info(json_data)
            scale_matches = [scale.key == key for scale in precomputed_metadata.scales]
            if not any(scale_matches):
                raise ValueError('Could not find key: {key} in info file at {info_path}' )
            else:
                scale_index = scale_matches.index(True)
                scale_meta = precomputed_metadata.scales[scale_index]
    else:
        scale_meta = ScaleMetadata(
            size=shape,
            resolution=resolution,
            encoding=encoding,
            chunk_size=chunks,
            key=key,
            voxel_offset=voxel_offset,
            jpeg_quality=jpeg_quality,
        )
        precomputed_metadata = PrecomputedMetadata(
            type=array_type, data_type=dtype, num_channels=num_channels, scales=[scale_meta]
        )

    if mode == "r":
        read = True
        # So cool that tensorstore errors when these are set to False for reading...
        write = None
        create = None
        delete_existing = None
    elif mode == 'a':
        read=True
        write=True
        create=True
        delete_existing=False
    elif mode == "rw":
        read = True
        write = True
        create = True
        delete_existing = True
    elif mode == "w":
        read = False
        write = True
        create = True
        delete_existing = True
    elif mode == "w-":
        read = False
        write = True
        create = True
        delete_existing = False
    else:
        raise ValueError('Mode must be "r", "rw", "a", "w", or "w-"')

    tsa = TensorStoreArray(
        driver=driver,
        path=_store_path,
        kvstore_path=kvstore_path,
        kvstore_driver=kvstore_driver,
        encoding=scale_meta.encoding,
        scale_index=scale_index,
        key=scale_meta.key,
        num_channels=precomputed_metadata.num_channels,
        volume_type=precomputed_metadata.type,
        dtype=precomputed_metadata.data_type,
        resolution=scale_meta.resolution,
        size=scale_meta.size,
        chunk_size=scale_meta.chunk_size,
        jpeg_quality=jpeg_quality
    )
    return tsa.open(
        read=read, write=write, create=create, delete_existing=delete_existing
    ).result()


def precomputed_to_dask(urlpath: str, chunks: Union[str, Sequence[int]], channel: int=0):
    store_path, key, _ = split_path_at_suffix(urlpath, ('.precomputed',))
    tsa = access_precomputed(store_path, key, mode='r')[ts.d["channel"][channel]]
    shape = tuple(tsa.shape)
    dtype = tsa.dtype.numpy_dtype
    if chunks == "original":
        chunks = tsa.spec().to_json()["scale_metadata"]["chunk_size"]
    _chunks = normalize_chunks(chunks, shape, dtype=dtype)
    def chunk_loader(store_path,key, block_info=None):
        idx = tuple(slice(*idcs) for idcs in block_info[None]["array-location"])
        tsa = access_precomputed(store_path, key, mode='r')[ts.d["channel"][channel]]
        result = tsa[idx].read().result()
        return result

    arr = map_blocks(chunk_loader, store_path, key, chunks=_chunks, dtype=dtype)
    return arr


@dataclass
class NicerTensorStore:
    spec: Dict[str, Any]
    open_kwargs: Dict[str, Any]

    def __getitem__(self, slices):
        return ts.open(spec=self.spec, **self.open_kwargs).result()[slices]

    def __setitem__(self, slices, values):
        ts.open(spec=self.spec, **self.open_kwargs).result()[ts.d["channel"][0]][
            slices
        ] = values
        return None


def prepare_tensorstore_from_pyramid(
    pyr: Sequence[DataArray],
    level_names: Sequence[str],
    jpeg_quality: int,
    output_chunks: Sequence[int],
    root_container_path: Path,
):
    store_arrays = []
    # sharding = {'@type': 'neuroglancer_uint64_sharded_v1',
    #       'preshift_bits': 9,
    #        'hash': 'identity',
    #        'minishard_index_encoding': 'gzip',
    #       'minishard_bits': 6,
    #       'shard_bits': 15}

    for p, ln in zip(pyr, level_names):
        res = [abs(float(p.coords[k][1] - p.coords[k][0])) for k in p.dims]
        spec: Dict[str, Any] = {
            "driver": "neuroglancer_precomputed",
            "kvstore": {
                "driver": "file",
                "path": str(Path(root_container_path).parent),
            },
            "path": root_container_path.parts[-1],
            "scale_metadata": {
                "size": p.shape,
                "resolution": res,
                "encoding": "jpeg",
                "jpeg_quality": jpeg_quality,
                #'sharding': sharding,
                "chunk_size": output_chunks,
                "key": ln,
                "voxel_offset": (0, 0, 0),
            },
            "multiscale_metadata": {
                "data_type": p.dtype.name,
                "num_channels": 1,
                "type": "image",
            },
        }
        try:
            ts.open(spec=spec, open=True).result()
        except ValueError:
            try:
                ts.open(spec=spec, create=True).result()
            except ValueError:
                ts.open(spec=spec, create=True, delete_existing=True).result()

        nicer_array = NicerTensorStore(spec=spec, open_kwargs={"write": True})
        store_arrays.append(nicer_array)
    return store_arrays



from .attrs import (VolumeStorageSpec, 
                   ContrastLimits, 
                   DisplaySettings, 
                   SpatialTransform, 
                   DatasetView, 
                   MultiscaleSpec, 
                   VolumeSource, 
                   DatasetIndex, 
                   VolumeIngest,
                   COSEMArrayAttrs,
                   NeuroglancerGroupAttrs,
                   MultiscaleGroupAttrs,
                   makeMultiscaleGroupAttrs,
                   CompositeArrayAttrs, 
                   MeshSource) 



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['fibsem_tools', 'fibsem_tools.attrs', 'fibsem_tools.io']

package_data = \
{'': ['*']}

install_requires = \
['dacite>=1.6.0,<2.0.0',
 'distributed>=2021.1.1,<2022.0.0',
 'fsspec>=0.8.4,<0.9.0',
 'h5py>=3.1.0,<4.0.0',
 'mrcfile>=1.2.0,<2.0.0',
 's3fs>=0.5.1,<0.6.0',
 'tensorstore>=0.1.8,<0.2.0',
 'zarr>=2.6.1,<3.0.0']

setup_kwargs = {
    'name': 'fibsem-tools',
    'version': '0.1.9',
    'description': 'Tools for processing FIBSEM datasets',
    'long_description': None,
    'author': 'Davis Vann Bennett',
    'author_email': 'davis.v.bennett@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""
Functions for reading FIB-SEM data from Harald Hess' proprietary format
Adapted from https://github.com/janelia-cosem/FIB-SEM-Aligner/blob/master/fibsem.py
"""

import os
import numpy as np
from typing import Iterable, Union, Sequence
from pathlib import Path
import dask.array as da
from xarray import DataArray

Pathlike = Union[str, Path]

# This value is used to ensure that the endianness of the data is correct
MAGIC_NUMBER = 3555587570
# This is the size of the header, in bytes. Everything beyond this number of bytes is data.
OFFSET = 1024


class FIBSEMHeader(object):
    """Structure to hold header info"""

    def __init__(self, **kwargs):
        self.__dict__ = kwargs

    def update(self, **kwargs):
        """update internal dictionary"""
        self.__dict__.update(kwargs)

    def to_native_types(self):
        for k, v in self.__dict__.items():
            if isinstance(v, np.integer):
                self.__dict__[k] = int(v)
            elif isinstance(v, np.bytes_):
                self.__dict__[k] = v.tostring().decode("utf-8")
            elif isinstance(v, np.ndarray):
                self.__dict__[k] = v.tolist()
            elif isinstance(v, np.floating):
                self.__dict__[k] = float(v)
            else:
                self.__dict__[k] = v


class FIBSEMData(np.ndarray):
    """Subclass of ndarray to attach header data to fibsem data"""

    def __new__(cls, input_array, header=None):
        # Input array is an already formed ndarray instance
        # We first cast to be our class type
        obj = np.asarray(input_array).view(cls)
        # add the new attribute to the created instance
        obj.header = header
        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None:
            return
        self.header = getattr(obj, "header", None)


class _DTypeDict(object):
    """Handle dtype dict manipulations"""

    def __init__(self, names=None, formats=None, offsets=None):
        # initialize internals to empty lists
        self.names = []
        self.formats = []
        self.offsets = []
        if names is not None:
            self.update(names, formats, offsets)

    def update(self, names, formats, offsets):
        """"""
        if isinstance(names, list):
            if len(names) == len(formats) == len(offsets):
                self.names.extend(names)
                self.formats.extend(formats)
                self.offsets.extend(offsets)
            else:
                raise RuntimeError("Lengths are not equal")
        else:
            self.names.append(names)
            self.formats.append(formats)
            self.offsets.append(offsets)

    @property
    def dict(self):
        """Return the dict representation"""
        return dict(names=self.names, formats=self.formats, offsets=self.offsets)

    @property
    def dtype(self):
        """return the dtype"""
        return np.dtype(self.dict)


def _read_header(path):
    # make emtpy header to fill
    header_dtype = _DTypeDict()
    header_dtype.update(
        [
            "FileMagicNum",  # Read in magic number, should be 3555587570
            "FileVersion",  # Read in file version number
            "FileType",  # Read in file type, 1 is Zeiss Neon detectors
            "SWdate",  # Read in SW date
            "TimeStep",  # Read in AI sampling time (including oversampling) in seconds
            "ChanNum",  # Read in number of channels
            "EightBit",  # Read in 8-bit data switch
        ],
        [">u4", ">u2", ">u2", ">S10", ">f8", ">u1", ">u1"],
        [0, 4, 6, 8, 24, 32, 33],
    )
    # read initial header
    with open(path, mode="rb") as fobj:
        base_header = np.fromfile(fobj, dtype=header_dtype.dtype, count=1)

        if len(base_header) == 0:
            raise RuntimeError(
                f"Base header is missing for {fobj.name}"
            )

        fibsem_header = FIBSEMHeader(
            **dict(zip(base_header.dtype.names, base_header[0]))
        )
        # now fobj is at position 34, return to 0
        fobj.seek(0, os.SEEK_SET)
        if fibsem_header.FileMagicNum != MAGIC_NUMBER:
            raise RuntimeError(
                f"FileMagicNum should be {MAGIC_NUMBER} but is {fibsem_header.FileMagicNum}"
            )

        _scaling_offset = 36
        if fibsem_header.FileVersion == 1:
            header_dtype.update(
                "Scaling", (">f8", (fibsem_header.ChanNum, 4)), _scaling_offset
            )
        elif fibsem_header.FileVersion in {2, 3, 4, 5, 6}:
            header_dtype.update(
                "Scaling", (">f4", (fibsem_header.ChanNum, 4)), _scaling_offset
            )
        else:
            # Read in AI channel scaling factors, (col#: AI#), (row#: offset, gain, 2nd order, 3rd order)
            header_dtype.update("Scaling", (">f4", (2, 4)), _scaling_offset)

        if fibsem_header.FileVersion >= 9:
            header_dtype.update(
                ["Restart", "StageMove", "FirstX", "FirstY"],
                [">u1", ">u1", ">i4", ">i4"],
                [68, 69, 70, 74],
            )

        header_dtype.update(
            ["XResolution", "YResolution"],  # X Resolution  # Y Resolution
            [">u4", ">u4"],
            [100, 104],
        )

        if fibsem_header.FileVersion in {1, 2, 3}:
            header_dtype.update(
                ["Oversampling", "AIDelay"],  # AI oversampling  # Read AI delay (
                [">u1", ">i2"],
                [108, 109],
            )
        else:
            header_dtype.update("Oversampling", ">u2", 108)  # AI oversampling

        header_dtype.update("ZeissScanSpeed", ">u1", 111)  # Scan speed (Zeiss #)
        if fibsem_header.FileVersion in {1, 2, 3}:
            header_dtype.update(
                [
                    "ScanRate",  # Actual AO (scanning) rate
                    "FramelineRampdownRatio",  # Frameline rampdown ratio
                    "Xmin",  # X coil minimum voltage
                    "Xmax",  # X coil maximum voltage
                ],
                [">f8", ">f8", ">f8", ">f8"],
                [112, 120, 128, 136],
            )
            # fibsem_header.Detmin = -10 # Detector minimum voltage
            # fibsem_header.Detmax = 10 # Detector maximum voltage
        else:
            header_dtype.update(
                [
                    "ScanRate",  # Actual AO (scanning) rate
                    "FramelineRampdownRatio",  # Frameline rampdown ratio
                    "Xmin",  # X coil minimum voltage
                    "Xmax",  # X coil maximum voltage
                    "Detmin",  # Detector minimum voltage
                    "Detmax",  # Detector maximum voltage
                    "DecimatingFactor",  # Decimating factor
                ],
                [">f4", ">f4", ">f4", ">f4", ">f4", ">f4", ">u2"],
                [112, 116, 120, 124, 128, 132, 136],
            )

        header_dtype.update(
            [
                "AI1",  # AI Ch1
                "AI2",  # AI Ch2
                "AI3",  # AI Ch3
                "AI4",  # AI Ch4
            ],
            [">u1", ">u1", ">u1", ">u1"],
            [151, 152, 153, 154],
        )

        if fibsem_header.FileVersion >= 9:
            header_dtype.update(
                ["SampleID"],
                ["S25"],
                [155],
            )

        header_dtype.update(
            ["Notes"],
            [">S200"],
            [180],
        )

        if fibsem_header.FileVersion in {1, 2}:
            header_dtype.update(
                [
                    "DetA",  # Name of detector A
                    "DetB",  # Name of detector B
                    "DetC",  # Name of detector C
                    "DetD",  # Name of detector D
                    "Mag",  # Magnification
                    "PixelSize",  # Pixel size in nm
                    "WD",  # Working distance in mm
                    "EHT",  # EHT in kV
                    "SEMApr",  # SEM aperture number
                    "HighCurrent",  # high current mode (1=on, 0=off)
                    "SEMCurr",  # SEM probe current in A
                    "SEMRot",  # SEM scan roation in degree
                    "ChamVac",  # Chamber vacuum
                    "GunVac",  # E-gun vacuum
                    "SEMStiX",  # SEM stigmation X
                    "SEMStiY",  # SEM stigmation Y
                    "SEMAlnX",  # SEM aperture alignment X
                    "SEMAlnY",  # SEM aperture alignment Y
                    "StageX",  # Stage position X in mm
                    "StageY",  # Stage position Y in mm
                    "StageZ",  # Stage position Z in mm
                    "StageT",  # Stage position T in degree
                    "StageR",  # Stage position R in degree
                    "StageM",  # Stage position M in mm
                    "BrightnessA",  # Detector A brightness (
                    "ContrastA",  # Detector A contrast (
                    "BrightnessB",  # Detector B brightness (
                    "ContrastB",  # Detector B contrast (
                    "Mode",
                    # FIB mode: 0=SEM, 1=FIB, 2=Milling, 3=SEM+FIB, 4=Mill+SEM, 5=SEM Drift Correction, 6=FIB Drift Correction, 7=No Beam, 8=External, 9=External+SEM
                    "FIBFocus",  # FIB focus in kV
                    "FIBProb",  # FIB probe number
                    "FIBCurr",  # FIB emission current
                    "FIBRot",  # FIB scan rotation
                    "FIBAlnX",  # FIB aperture alignment X
                    "FIBAlnY",  # FIB aperture alignment Y
                    "FIBStiX",  # FIB stigmation X
                    "FIBStiY",  # FIB stigmation Y
                    "FIBShiftX",  # FIB beam shift X in micron
                    "FIBShiftY",  # FIB beam shift Y in micron
                ],
                [
                    ">S10",
                    ">S18",
                    ">S20",
                    ">S20",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">u1",
                    ">u1",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">u1",
                    ">f8",
                    ">u1",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                    ">f8",
                ],
                [
                    380,
                    390,
                    700,
                    720,
                    408,
                    416,
                    424,
                    432,
                    440,
                    441,
                    448,
                    456,
                    464,
                    472,
                    480,
                    488,
                    496,
                    504,
                    512,
                    520,
                    528,
                    536,
                    544,
                    552,
                    560,
                    568,
                    576,
                    584,
                    600,
                    608,
                    616,
                    624,
                    632,
                    640,
                    648,
                    656,
                    664,
                    672,
                    680,
                ],
            )
        else:
            header_dtype.update(
                [
                    "DetA",  # Name of detector A
                    "DetB",  # Name of detector B
                    "DetC",  # Name of detector C
                    "DetD",  # Name of detector D
                    "Mag",  # Magnification
                    "PixelSize",  # Pixel size in nm
                    "WD",  # Working distance in mm
                    "EHT",  # EHT in kV
                    "SEMApr",  # SEM aperture number
                    "HighCurrent",  # high current mode (1=on, 0=off)
                    "SEMCurr",  # SEM probe current in A
                    "SEMRot",  # SEM scan roation in degree
                    "ChamVac",  # Chamber vacuum
                    "GunVac",  # E-gun vacuum
                    "SEMShiftX",  # SEM beam shift X
                    "SEMShiftY",  # SEM beam shift Y
                    "SEMStiX",  # SEM stigmation X
                    "SEMStiY",  # SEM stigmation Y
                    "SEMAlnX",  # SEM aperture alignment X
                    "SEMAlnY",  # SEM aperture alignment Y
                    "StageX",  # Stage position X in mm
                    "StageY",  # Stage position Y in mm
                    "StageZ",  # Stage position Z in mm
                    "StageT",  # Stage position T in degree
                    "StageR",  # Stage position R in degree
                    "StageM",  # Stage position M in mm
                    "BrightnessA",  # Detector A brightness (#)
                    "ContrastA",  # Detector A contrast (#)
                    "BrightnessB",  # Detector B brightness (#)
                    "ContrastB",  # Detector B contrast (#)
                    "Mode",
                    # FIB mode: 0=SEM, 1=FIB, 2=Milling, 3=SEM+FIB, 4=Mill+SEM, 5=SEM Drift Correction, 6=FIB Drift Correction, 7=No Beam, 8=External, 9=External+SEM
                    "FIBFocus",  # FIB focus in kV
                    "FIBProb",  # FIB probe number
                    "FIBCurr",  # FIB emission current
                    "FIBRot",  # FIB scan rotation
                    "FIBAlnX",  # FIB aperture alignment X
                    "FIBAlnY",  # FIB aperture alignment Y
                    "FIBStiX",  # FIB stigmation X
                    "FIBStiY",  # FIB stigmation Y
                    "FIBShiftX",  # FIB beam shift X in micron
                    "FIBShiftY",  # FIB beam shift Y in micron
                ],
                [
                    ">S10",
                    ">S18",
                    ">S20",
                    ">S20",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">u1",
                    ">u1",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">u1",
                    ">f4",
                    ">u1",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                ],
                [
                    380,
                    390,
                    410,
                    430,
                    460,
                    464,
                    468,
                    472,
                    480,
                    481,
                    490,
                    494,
                    498,
                    502,
                    510,
                    514,
                    518,
                    522,
                    526,
                    530,
                    534,
                    538,
                    542,
                    546,
                    550,
                    554,
                    560,
                    564,
                    568,
                    572,
                    600,
                    604,
                    608,
                    620,
                    624,
                    628,
                    632,
                    636,
                    640,
                    644,
                    648,
                ],
            )

        if fibsem_header.FileVersion in {5, 6, 7, 8}:
            header_dtype.update(
                [
                    "MillingXResolution",  # FIB milling X resolution
                    "MillingYResolution",  # FIB milling Y resolution
                    "MillingXSize",  # FIB milling X size (um)
                    "MillingYSize",  # FIB milling Y size (um)
                    "MillingULAng",  # FIB milling upper left inner angle (deg)
                    "MillingURAng",  # FIB milling upper right inner angle (deg)
                    "MillingLineTime",  # FIB line milling time (s)
                    "FIBFOV",  # FIB FOV (um)
                    "MillingLinesPerImage",  # FIB milling lines per image
                    "MillingPIDOn",  # FIB milling PID on
                    "MillingPIDMeasured",  # FIB milling PID measured (0:specimen, 1:beamdump)
                    "MillingPIDTarget",  # FIB milling PID target
                    "MillingPIDTargetSlope",  # FIB milling PID target slope
                    "MillingPIDP",  # FIB milling PID P
                    "MillingPIDI",  # FIB milling PID I
                    "MillingPIDD",  # FIB milling PID D
                    "MachineID",  # Machine ID
                    "SEMSpecimenI",  # SEM specimen current (nA)
                ],
                [
                    ">u4",
                    ">u4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">u2",
                    ">u1",
                    ">u1",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">f4",
                    ">S30",
                    ">f4",
                ],
                [
                    652,
                    656,
                    660,
                    664,
                    668,
                    672,
                    676,
                    680,
                    684,
                    686,
                    689,
                    690,
                    694,
                    698,
                    702,
                    706,
                    800,
                    980,
                ],
            )

        if fibsem_header.FileVersion in {6, 7}:
            header_dtype.update(
                [
                    "Temperature",  # Temperature (F)
                    "FaradayCupI",  # Faraday cup current (nA)
                    "FIBSpecimenI",  # FIB specimen current (nA)
                    "BeamDump1I",  # Beam dump 1 current (nA)
                    "SEMSpecimenI",  # SEM specimen current (nA)
                    "MillingYVoltage",  # Milling Y voltage (V)
                    "FocusIndex",  # Focus index
                    "FIBSliceNum",  # FIB slice #
                ],
                [">f4", ">f4", ">f4", ">f4", ">f4", ">f4", ">f4", ">u4"],
                [850, 854, 858, 862, 866, 870, 874, 878],
            )
        if fibsem_header.FileVersion == 8:
            header_dtype.update(
                [
                    "BeamDump2I",  # Beam dump 2 current (nA)
                    "MillingI",  # Milling current (nA)
                ],
                [">f4", ">f4"],
                [882, 886],
            )
        header_dtype.update("FileLength", ">i8", 1000)  # Read in file length in bytes

        # read header
        header = np.fromfile(fobj, dtype=header_dtype.dtype, count=1)
        fibsem_header = FIBSEMHeader(**dict(zip(header.dtype.names, header[0])))
        fibsem_header.to_native_types()

    return fibsem_header


def _read(path: Union[str, Path]) -> FIBSEMData:
    """

    Read a single .dat file.

    Parameters
    ----------
    path : string denoting a path to a .dat file

    Returns : an instance of FIBSEMData
    -------

    """
    # Load raw_data data file 's' or 'ieee-be.l64' Big-ian ordering, 64-bit long data type
    fibsem_header = _read_header(path)
    # read data
    shape = (
        fibsem_header.YResolution,
        fibsem_header.XResolution,
        fibsem_header.ChanNum,
    )

    if fibsem_header.EightBit == 1:
        dtype = ">u1"
    else:
        dtype = ">i2"

    try:
        raw_data = np.memmap(
            path,
            dtype=dtype,
            mode="r",
            offset=OFFSET,
            shape=shape,
        )
    except ValueError:
        # todo: read what data is available
        raw_data = np.zeros(dtype=dtype, shape=shape)

    raw_data = np.rollaxis(raw_data, 2)
    # Once read into the FIBSEMData structure it will be in memory, not memmap.
    result = FIBSEMData(raw_data, fibsem_header)
    del raw_data
    return result


def read_fibsem(path: Union[str, Path, Iterable[str], Iterable[Path]]):
    """

    Parameters
    ----------
    path : string or iterable of strings representing paths to .dat files.

    Returns a single FIBSEMDataset or an iterable of FIBSEMDatasets, depending on whether a single path or multiple
    paths are supplied as arguments.
    -------

    """
    if isinstance(path, (str, Path)):
        return _read(path)
    elif isinstance(path, Iterable):
        return [_read(p) for p in path]
    else:
        raise ValueError(
            f"Path '{path}' must be an instance of string or pathlib.Path, or iterable of strings / pathlib.Paths"
        )


def _convert_data(fibsem):
    """

    Parameters
    ----------
    fibsem

    Returns
    -------

    """
    # Convert raw_data data to electron counts
    if fibsem.header.EightBit == 1:
        scaled = np.empty(fibsem.shape, dtype=np.int16)
        detector_a, detector_b = fibsem
        if fibsem.header.AI1:
            detector_a = fibsem[0]
            scaled[0] = np.int16(
                fibsem[0]
                * fibsem.header.ScanRate
                / fibsem.header.Scaling[0, 0]
                / fibsem.header.Scaling[0, 2]
                / fibsem.header.Scaling[0, 3]
                + fibsem.header.Scaling[0, 1]
            )
            if fibsem.header.AI2:
                detector_b = fibsem[1]
                scaled[1] = np.int16(
                    fibsem[1]
                    * fibsem.header.ScanRate
                    / fibsem.header.Scaling[1, 0]
                    / fibsem.header.Scaling[1, 2]
                    / fibsem.header.Scaling[1, 3]
                    + fibsem.header.Scaling[1, 1]
                )

        elif fibsem.header.AI2:
            detector_b = fibsem[0]
            scaled[0] = np.int16(
                fibsem[0]
                * fibsem.header.ScanRate
                / fibsem.header.Scaling[0, 0]
                / fibsem.header.Scaling[0, 2]
                / fibsem.header.Scaling[0, 3]
                + fibsem.header.Scaling[0, 1]
            )

    else:
        if fibsem.header.FileVersion in {1, 2, 3, 4, 5, 6}:
            # scaled =

            if fibsem.header.AI1:
                detector_a = (
                    fibsem.header.Scaling[0, 0]
                    + fibsem[0] * fibsem.header.Scaling[0, 1]
                )
                if fibsem.header.AI2:
                    detector_b = (
                        fibsem.header.Scaling[1, 0]
                        + fibsem[1] * fibsem.header.Scaling[1, 1]
                    )
                    if fibsem.header.AI3:
                        detector_c = (
                            fibsem.header.Scaling[2, 0]
                            + fibsem[1] * fibsem.header.Scaling[1, 1]
                        )
        else:
            pass
    return detector_a, detector_b, scaled


def _chunked_fibsem_loader(
    filenames, channel_axis, pad_values=None, concat_axis=0, block_info=None
):
    """
    Load fibsem data and pad if needed. Designed to work with da.map_blocks.
    """
    idx = block_info[None]["chunk-location"]
    output_shape = block_info[None]["chunk-shape"]
    filedata = np.expand_dims(
        np.asanyarray(read_fibsem(filenames[idx[0]])), concat_axis
    )
    pad_width = np.subtract(output_shape, filedata.shape)
    if np.any(pad_width):
        if not pad_values:
            raise ValueError("Data must be padded but no pad values were supplied!")
        padded = []
        pw = np.take(pad_width, [x for x in range(len(pad_width)) if x != channel_axis])
        for ind, channel in enumerate(np.swapaxes(filedata, 0, channel_axis)):
            padded.append(
                np.pad(
                    channel,
                    list(zip(pw * 0, pw)),
                    mode="constant",
                    constant_values=pad_values[ind],
                )
            )

        result = np.stack(padded, channel_axis)
    else:
        result = filedata
    return result


def minmax(filenames, block_info):
    idx = block_info[None]["chunk-location"][0]
    filedata = read_fibsem(filenames[idx])
    return np.expand_dims(np.array([[f.min(), f.max()] for f in filedata]), 0)


def aggregate_fibsem_metadata(fnames):
    headers = [_read_header(f) for f in fnames]
    meta, shapes, dtypes = [], [], []
    # build lists of metadata for each image
    for d in headers:
        if d.EightBit == 1:
            dtype = ">u1"
        else:
            dtype = ">i2"
        meta.append(d.__dict__)
        shapes.append((d.ChanNum, d.YResolution, d.XResolution))
        dtypes.append(dtype)
    return meta, shapes, dtypes


class FibsemDataset:
    def __init__(self, filenames: Sequence[str]):
        """
        Create a representation of a collection of .dat files as a single dataset.
        """
        self.filenames = filenames
        self.metadata, self.shapes, self.dtypes = aggregate_fibsem_metadata(
            self.filenames
        )
        self.axes = {"z": 0, "c": 1, "y": 2, "x": 3}
        self.bounding_shape = {
            k: v
            for k, v in zip(
                self.axes, (len(self.filenames), *np.array(self.shapes).max(0))
            )
        }
        self.extrema = self._get_extrema()
        self.needs_padding = len(set(self.shapes)) > 1
        self.grid_spacing, self.coords = self._get_grid_spacing_and_coords()

    def _get_grid_spacing_and_coords(self):
        lateral = self.metadata[0]["PixelSize"]
        axial = (
            abs((self.metadata[0]["WD"] - self.metadata[-1]["WD"]) / len(self.metadata))
            / 1e-6
        )
        grid_spacing = {"z": axial, "c": 1, "y": lateral, "x": lateral}
        coords = {
            k: np.arange(0, grid_spacing[k] * self.bounding_shape[k], grid_spacing[k])
            for k in grid_spacing
        }
        return grid_spacing, coords

    def _get_extrema(self):
        all_extrema = da.map_blocks(
            minmax,
            self.filenames,
            chunks=((1,) * len(self.filenames), self.bounding_shape["c"], 2),
            dtype=self.dtypes[0],
        )
        result_data = da.stack([all_extrema.min(0)[:, 0], all_extrema.max(0)[:, 1]])
        result = DataArray(
            result_data, dims=("statistic", "channel"), coords={"statistic": ["min", "max"], 'channel': [0,1]}
        )
        return result

    def to_dask(self, pad_values=None):
        num_channels = self.bounding_shape["c"]
        if self.needs_padding:
            if pad_values is None:
                raise ValueError("Data must be padded but no pad values were supplied!")
            elif len(pad_values) != num_channels:
                raise ValueError(
                    f"Length of pad values {pad_values} does not match the length of the channel axis ({num_channels})"
                )

        chunks = (
            (1,) * self.bounding_shape["z"],
            *[self.bounding_shape[k] for k in ("c", "y", "x")],
        )
        darr = da.map_blocks(
            _chunked_fibsem_loader,
            self.filenames,
            self.axes["c"],
            pad_values,
            chunks=chunks,
            dtype=self.dtypes[0],
        )
        return darr
